/************************************************************************/
/*                                                                      
  	BANCOMM bc63{57}VME Time and Frequency Processor Device Driver.

  		For Sun Microsystems SunOS 5.3 and SunOS 4.1.3. 
                                                                      
	Copyright (C) 1994, BANCOMM, San Jose, CA.  All rights reserved.	
                                                                      
  	Author:	Steve Schoggen

		S. G. Schoggen & Company
		P. O. Box 246
		Little Silver, NJ  07739-0246	    908-530-4659

		Winter, 1994

	Editor tabstop settings:
		btfpu.h		tabstop=8
		btfpd.h		tabstop=8
		btfp.c		tabstop=8	(up to UNIX entry points)
				tabstop=4	(after UNIX entry points)
		btfptest.c	tabstop=4

	01/21/94..SGS..This file compiles on SunOS5.x and SunOS4.x.
			Use the Makefile to build for SunOS5; build
			the driver into the kernel for SunOS4.

			The 'wrappers' for each OS are grouped together
			and separate from each other, as these are
			vastly different between the two SunOS's.
			For example, there are separate probe and attach
			routines for SunOS4 and SunOS5.
			Thereafter, the UNIX entry points have the code
			#ifdef'ed to allow dual compilation.
			
			Note that "signal mode" is not supported in the
			SunOS5 version, only on the SunOS4 version.
			I could not find out how to send SIGUSR1 to
			the application from the interrupt routine.
			If this can be done, then just #define SIGNALS
			to enable signal mode again.
*/                                                                      
/****************************************************************************/
#include "btfpu.h"			/* Used by driver and applications.*/
#include "btfpd.h"			/* Used only by the driver.*/

/****************************************************************************/
/* Debugging message logger to conslole...
*/
/*
 * Defining DEBUG on the compile line (-DDEBUG) will enable debugging
 * statements in this driver.
 * Error message control:
 * 0 = no messages.
 * 1 = Error messages.
 * Can set the value of the debug flag using adb or in /etc/system file
 * with "set btfp:btfpdebug=<value>
 */

#ifdef  DEBUG
static  int     btfpdebug = 1;
#else
static  int     btfpdebug = 0;
#endif  DEBUG

/****************************************************************************/
/*	bc63{57} device operating mode structure. 
*/
static unsigned char mode1[5]=
{
			SOH,
			'A', 
			'1',	/* Mode 1 (Free Running Mode)*/
			ETB, 
			NULL
};

/************************************************************************/
/************************************************************************/
#ifdef sunos5
/* Start of SunOS5 Loadable Driver 'Wrapper' Code...
*/
/*---------------------------------------------------------------------------*/
static void *btfpstatep;        		/* Ptr to instance state struc*/

/*---------------------------------------------------------------------------*/
/* Function prototypes...
*/
static int btfpidentify( dev_info_t *dip );
static int btfpprobe(    dev_info_t *dip );
static int btfpattach(   dev_info_t *dip, ddi_attach_cmd_t cmd );
static int btfpdetach(   dev_info_t *dip, ddi_detach_cmd_t cmd );
static int btfpgetinfo(  dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
                                                                void **result );

static int btfpopen(     dev_t *devp, int flag, int otyp, cred_t *credp );
static int btfpclose(    dev_t   dev, int flag, int otyp, cred_t *credp);
static int btfpread(     dev_t   dev, struct uio *uiop,   cred_t *credp );
static int btfpwrite(    dev_t   dev, struct uio *uiop,   cred_t *credp );
static int btfpioctl(    dev_t   dev, int cmd,  int arg,  int flag, 
						cred_t *credp, int *rvalp );
static 
unsigned int btfpintr(   caddr_t instance );


/*---------------------------------------------------------------------------*/
/* Essential Data Structures...
*/
/* /// Is the required??
*/
struct bus_ops btfpbus_ops = {
        nullbusmap,
        0,              /* ddi_intrspec_t       (*bus_get_intrspec)(); */
        0,              /* int                  (*bus_add_intrspec)(); */
        0,              /* void                 (*bus_remove_intrspec)(); */
        0,              /* int                  (*bus_map_fault)() */
        0,              /* int                  (*bus_dma_map)() */
        0,              /* int                  (*bus_dma_ctl)() */
        ddi_ctlops,
        ddi_bus_prop_op,
};
 
/*
 * Device driver ops vector
 */
 
static struct cb_ops btfpcb_ops = {
 
        btfpopen,               /* open */
        btfpclose,              /* close */
        nodev,                  /* strategy */
        nodev,                  /* print */
        nodev,                  /* dump */
        btfpread,               /* read */
        btfpwrite,              /* write */
        btfpioctl,              /* ioctl */
        nodev,                  /* devmap */
        nodev,                  /* mmap */
        nodev,                  /* segmap */
        nochpoll,               /* poll */
        ddi_prop_op,            /* cb_prop_op */
        0,                      /* streamtab  */
#ifdef OMIT
        D_NEW  			/* Driver comaptibility flag- if using splx.*/
#endif OMIT
        D_NEW | D_MTSAFE        /* Driver comaptibility flag */
 
};
 
struct dev_ops btfp_dev_ops = {
        DEVO_REV,               /* devo_rev, */
        0,                      /* refcnt  */
        btfpgetinfo,            /* info */
        btfpidentify,           /* identify */
        btfpprobe,              /* probe */
        btfpattach,             /* attach */
        btfpdetach,             /* detach */
        nodev,                  /* reset */
        &btfpcb_ops,            /* driver operations */
        &btfpbus_ops,           /* bus operations */
};
 
/*---------------------------------------------------------------------------*/
extern struct mod_ops mod_driverops;
 
/*
 * Module linkage information for the kernel.
 */
static struct modldrv modldrv = 
{
        &mod_driverops, 			/* Adrs of std driver ops.*/
        "BANCOMM bc635VME device driver",	/* Returned by modinfo cmd.*/
        &btfp_dev_ops,       			/* This driver's ops.*/
};
 
static struct modlinkage modlinkage = 
{
        MODREV_1, 				/* Loadable modules system rev*/
        &modldrv,                               /* Null terminated list of */
        NULL                                    /* linkage structures.*/
};
 
/*--------------------------------------------------------------------------*/
/* Loadable driver support routines...
*/
/* Called by the 'add_drv' command...
*/
int
_init(void)                                     /* Load the driver...*/
{
        int error;                              /* Return value holder.*/
 
        SHOW( "btfpinit:   Called!\n" );
 
        if( ( error = ddi_soft_state_init( &btfpstatep, /* Init state ptr and */
            sizeof( struct btfp_status ), 1 ) ) != 0) /* struct size with sys.*/
        {
                return( error );
        }
 
        if( ( error = mod_install( &modlinkage ) ) != 0 ) /* Install driver.*/
        {
                ddi_soft_state_fini( &btfpstatep );  /* De-install on error.*/
        }
 
        /* Get per module resources as required.*/
        /* Do other device and/or driver initialization.*/
 
        return( error );                        /* Return val from mod_install*/
}

/*--------------------------------------------------------------------------*/
/* Called by the 'modinfo' command...
*/
int
_info( struct modinfo *modinfop )               /* Report information.*/
{
        SHOW( "btfp_info:  Called!\n" );
 
        return( mod_info( &modlinkage, modinfop ) ); /* Std for all drivders.*/
}
 
/*---------------------------------------------------------------------------*/
/* Called by the 'rem_drv' command...
*/
int
_fini( void )                                   /* Unload the driver...*/
{
        int status;                             /* Returned status holder.*/
 
        SHOW( "btfpfini:   Called!\n" );
 
        if( ( status = mod_remove( &modlinkage ) ) != 0 ) /* Unload driver.*/
        {
                cmn_err( CE_CONT,
                "btfpfini:   Error on mod_remove: %d!\n", status );
                goto exit;
        }
 
        /* Release, here, per-module resources as required!!*/
 
        ddi_soft_state_fini( &btfpstatep );       /* What does this do?? */
 
  exit:
        return( status );
}

/*---------------------------------------------------------------------------*/
/* Autoconfiguration Support Routines...
*/
/* Device identification.
	Note that the /usr/kernel/drv/btfp.conf file MUST indicate
	a device with name 'btfp'!!
*/
static int
btfpidentify(dev_info_t *dev)			/* Device identificaion.*/
{
        char *name = ddi_get_name(dev);
 
        SHOW( "btfpident...Called!\n" );
        if (strcmp(name, "btfp") == 0) {
                return (DDI_IDENTIFIED);
        } else {
                return (DDI_NOT_IDENTIFIED);
        }
}

/*---------------------------------------------------------------------------*/
/*	btfpprobe() - This routine is called at system boot or add_drv time.
		      To 'probe if the module is there', we read the bc63{57}'s 
		      ID and type registers.  If these reads do not cause bus 
		      errors, and if there values are 0xfef4 and 0xf350
		      respectively, then we assume a bc63{57} is there and 
		      return OK, else return probe failed. 

		      This routine could be enhanced to do other reads and
		      writes to the device to make a more thorough test, but
		      that probably is not necessary.  To test the device,
		      run the supplied 'btfptest' program after installation.
*/
static int
btfpprobe( dev_info_t *dip )
{
	struct btfp_device *dp;			/* Pointer to device regs.*/
        unsigned short id, typ;                 /* To read device id and type.*/
 
 
        SHOW( "btfpprobe:  Called!\n" );
 
						/* Map in device registers.*/
        if( ddi_map_regs( dip, (u_int)0, (caddr_t*)&dp, (off_t)0,
					(off_t)sizeof( struct btfp_device ) ) )
        {
                cmn_err( CE_CONT,
                "btfpprobe:  Unable to map registers for btfp0.");
                goto failure;
        }
 
        					/* Read board ID register.*/ 
        if( ddi_peeks( dip, (short*)&dp->btfp_id, &id ) != DDI_SUCCESS )
        {
                cmn_err( CE_CONT, "%s%d:      Cannot read device bc635 ID.",
                         ddi_get_name(dip), ddi_get_instance(dip));
                goto failure;
        }
        else
                if( btfpdebug ) 
		{
			cmn_err( CE_CONT, "%s%d:      Device ID is   0x%x\n",
                        ddi_get_name(dip), ddi_get_instance(dip), id );
		}
 
        					/* Read board ID register.*/ 
        if( ddi_peeks( dip, (short*)&dp->btfp_type, &typ ) != DDI_SUCCESS )
        {
                cmn_err( CE_CONT, "%s%d:      Cannot read device type.",
                         ddi_get_name(dip), ddi_get_instance(dip));
                goto failure;
        }
        else
                if( btfpdebug ) 
		{
			cmn_err( CE_CONT, "%s%d:      Device type is 0x%x\n",
                        ddi_get_name(dip), ddi_get_instance(dip), typ );
		}
 
        if( id != BC635ID || typ != BC635TYPE ) /* This is not a bc635!!*/
        {
                cmn_err( CE_WARN,
                "The module configured as %s%d is not a BANCOMM bc635!\n",
                         ddi_get_name(dip), ddi_get_instance(dip) );
                goto failure;                   /* Exit on ID error.*/
        }
 
success:
        SHOW( "btfpprobe:  Returning Success!\n" );
        if( dp )                                /* Unmap regs and return.*/
                ddi_unmap_regs( dip, 0, (caddr_t*)&dp, (off_t) 0, 
					(off_t)sizeof( struct btfp_device ) );
        return (DDI_PROBE_SUCCESS);
 
failure:                                     /* Unmap regs if mapped; exit.*/
        SHOW( "btfpprobe:  Returning Failure!\n" );
        if( dp )
                ddi_unmap_regs( dip, 0, (caddr_t*)&dp, (off_t) 0,
					(off_t)sizeof( struct btfp_device ) );
        return (DDI_PROBE_FAILURE);
}

/*---------------------------------------------------------------------------*/
/*      btfpattach() - This routine is called at system boot time.
                        It initializes the bc63{57}'s interrupt vector and
                        priority level, and sets the device status structure's
                        interrupt indication to not waiting for an interrupt.
                        (The module comes up on system reset with interrupts
                        disabled.)
			Note:  This whole routine is SunOS5 code only.
				See the separate SunOS4 attach routine.
*/       

static int
btfpattach( dev_info_t *dip, ddi_attach_cmd_t cmd )
{
	struct btfp_device *dp;			/* Pointer to device regs.*/
        struct btfp_status *sp;                 /* Ptr to state/status struct.*/
	struct interruptinfo *iip;		/* Interrupt vector/level prt.*/
        int    instance;                        /* Returned from system.*/
        char   namesfx[ 3 ];                    /* Name suffix for minor node.*/

        SHOW( "btfpattach: Called!\n" );

        switch ( cmd )
        {
        case DDI_ATTACH:
                instance = ddi_get_instance( dip ); /* Instance from system.*/
                                                /* Allocate state structure.*/
                if( ddi_soft_state_zalloc( btfpstatep, instance ) != 0 )
                {
                        cmn_err(CE_CONT,
				"%s%d:      Can't allocated state struct.\n",
						ddi_get_name( dip ), instance );
                        return( DDI_FAILURE );
                }
                else
                {
                        if( btfpdebug ) cmn_err( CE_CONT, 
				"%s%d:      Allocated state struct OK!\n",
                        			ddi_get_name( dip ), instance );
                }
 
                sp = ddi_get_soft_state( btfpstatep, instance ); /* Ptr 2 same*/

                /* Init softstate stuff here.*/
                sp->dip = dip;                  /* Save info handle for later.*/
                sp->intwait = 0;        	/* Set not waiting for an int.*/
                sp->signalon  = 0;      	/* Signal mechanism turned off*/
                sp->packeton  = 0;      	/* Packet mode is turned off.*/
                sp->debugon  = 0;       	/* Debug mode is turned off.*/
                sp->gps_synch  = 0;     	/* Not synched with GPS pkts.*/

                /* Register interrupts here....*/
		if( ddi_intr_hilevel( dip, 0 ) != 0 ) /* Don't support these.*/
		{
			cmn_err( CE_CONT, 
		     "btfpattach: High-level interrupts are not supported!\n" );
			return( DDI_FAILURE );
		}
         	/* Interrupt number is 0 (we only use one interrupt level)
         	   A magic cookie for later use in MUTEXes will be returned 
		   in sp->iblock_cookie.  Pass instance to btfpintr().
		   But first, a null handler is required..see p101 of SunOS5
		   driver manual to explain this nonsense...	
         	*/
        	if( ddi_add_intr( dip, 0, &sp->iblock_cookie,
            				  NULL, ( u_int(*)(caddr_t) )nulldev,
            				  NULL ) != DDI_SUCCESS )
               		return( DDI_FAILURE );

		/* Initialize mutual exclusion lock here...*/
        	mutex_init( &sp->mutex, "btfp mutex", MUTEX_DRIVER, 
						(void *)sp->iblock_cookie );

		ddi_remove_intr( dip, 0, sp->iblock_cookie );
						/* Now the real intr handler.*/	
        	if( ddi_add_intr( dip, 0, &sp->iblock_cookie,
            				  &sp->idevice_cookie, btfpintr,
            				  (caddr_t)instance ) != DDI_SUCCESS )
               		return( DDI_FAILURE );

		/* Initialize mutual condition variable here...*/
		cv_init( &sp->cv, "btfp cv", CV_DRIVER, NULL );

               	/* Map device registers to pointer in soft-state structure.*/
        	if( ddi_map_regs( dip, (u_int)0, (caddr_t*)&sp->regp, (off_t)0,
                                        (off_t)sizeof( struct btfp_device ) ) )
        	{
                	cmn_err( CE_CONT,
                	"btfpattach: Unable to map registers for %s%d.",
                        			ddi_get_name( dip ), instance );
                        ( void )btfpdetach( dip, DDI_DETACH );
        	}
		dp = sp->regp;			/* Working device regs pointer*/
						/* See that mapping worked...*/
		if( btfpdebug )
			cmn_err( CE_CONT, "btfpattach: Device type is 0x%x\n", 
								dp->btfp_type );

                /* Program device interrupt level and vector numbers.*/
		iip = (struct interruptinfo*)&sp->idevice_cookie;
		dp->btfp_vector = iip->vector;  /* Program VME int vector.*/
		dp->btfp_level  = iip->priority;/* Program VME int priority.*/
		if( btfpdebug )
			cmn_err( CE_CONT,
			"btfpattach: idevice_cookie 0x%x  vec 0x%x  pri 0x%x\n",
			sp->idevice_cookie,iip->vector,iip->priority);

                /* Create minor node...*/
		/* ///Can not find a 'ddi_character' type, but this works!? */
                sprintf( namesfx, "%d", instance); /* Instance# to string.*/
                if( ddi_create_minor_node( dip, namesfx, S_IFCHR, instance, 
					DDI_PSEUDO, NULL ) == DDI_FAILURE )
                {
                                                /* Use our own detach routine */
                                                /* to toss away any stuff we  */
                                                /* allocated above.*/
                        cmn_err(CE_CONT,
				"%s%d:      ddi_create_minor_node failed!\n",
                        			ddi_get_name( dip ), instance );
                        ( void )btfpdetach( dip, DDI_DETACH );
                        return( DDI_FAILURE );
                }
 
                ddi_report_dev( dip );          /* Log device name on console.*/
                SHOW( "btfpattach: Success-attach completed OK!\n" );
                return( DDI_SUCCESS );          /* Attach succeeded!!*/
                break;
 
        default:
                return( DDI_FAILURE );          /* Support attach only.*/
        }
 
};
 
/*--------------------------------------------------------------------------*/
static int
btfpdetach( dev_info_t *dip, ddi_detach_cmd_t cmd )
{
	struct btfp_device *dp;			/* Pointer to device regs.*/
        struct btfp_status *sp;                 /* Pointer to softstate struc.*/
        int    instance;                        /* Device instance from system*/
        
        SHOW( "btfpdetach: Called!\n" );
 
        switch( cmd )
        {
        case DDI_DETACH:
                instance = ddi_get_instance( dip ); /* Get our instance #.*/
                sp = ddi_get_soft_state( btfpstatep, instance ); /* Struct ptr*/
		dp = sp->regp;			/* Pointer to device regs.*/
 
                /* Put device specific logic here- make quiescent, etc...*/
                /* e.g. turn off device and it's interrupts.*/
                /* Remove interrupt, mutexes, and properties here.*/
                /* Unmap regs here.*/
 
                ddi_remove_minor_node( dip, NULL );       /* Remove device.*/
		ddi_remove_intr( dip, 0, sp->iblock_cookie ); /* Interrupt.*/
		ddi_unmap_regs( dip, 0, (caddr_t*)&dp, (off_t) 0,
                                        (off_t)sizeof( struct btfp_device ) );
		cv_destroy( &sp->cv );		/* Destroy condition variable*/
		mutex_destroy( &sp->mutex );	/* Destroy mutex lock.*/
                ddi_soft_state_free(btfpstatep,instance); /* Free state struc*/ 
                                                          /* Do this last.*/
 
                SHOW( "btfpdetach: Returning success!\n" );

                return( DDI_SUCCESS );          /* Detach succeeded!!*/
 
        default:
                SHOW( "btfpdetach: Returning FAILURE!\n" );
                return( DDI_FAILURE );          /* Support detach call only.*/
        }
}
 
/*--------------------------------------------------------------------------*/
/* Return to system, info known only to the driver...
        Manual says driver must not assume state struct is available, as
        getinfo can be called before attach.
*/
static int
btfpgetinfo( dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result )
{
        struct btfp_status *sp;                 /* State structure pointer.*/
        int error = DDI_FAILURE;                /* Default to error return.*/


        SHOW( "btfpgetinfo:Called!\n" );

        switch( infocmd )
        {
        case DDI_INFO_DEVT2DEVINFO:             /* Asks for pointer to devinfo*/                if( ( sp = ddi_get_soft_state( btfpstatep,
                                getminor( (dev_t)arg ) ) ) != NULL )
                {
                        *result = sp->dip;
                        error = DDI_SUCCESS;
                }
                else
                {
                        *result = NULL;
                }
                break;
 
        case DDI_INFO_DEVT2INSTANCE:            /* Asks for device instance #.*/                *result = (void *)getminor( (dev_t) arg); /* Get minor dev#.*/
                error = DDI_SUCCESS;            /* No problem.*/
                break;
 
        default:
                break;
        }
 
        return (error);
}
 
 
#endif sunos5
/* End of SunOS5 loadable/autoconfiguration wrapper.
	(Note: ':set tabstop=8' above and to this point)
*/
/*****************************************************************************/
/*****************************************************************************/



/*****************************************************************************/
/*****************************************************************************/
/* Start of SunOS4 autoconfiguration wrapper (SunOS4 driver is not loadable)...
*/
/*****************************************************************************/
#ifdef sunos4
/*
    SUN(os4) Autoconfiguration-related declarations...
    Autoconfiguration is done at system boot time.
*/
int btfpprobe();                    /* Kernel interface routines called by  */
int btfpattach();                   /* autoconfig for this driver.          */

struct mb_device *btfpinfo[ NBTFP ]; /* Initialized by 'config' for each dev.*/

struct mb_driver btfpdriver = { btfpprobe, 0, btfpattach, 0, 0, 0,
                                    sizeof( struct btfp_device ),
                                    "btfp", btfpinfo, 0, 0, 0, 0 };

struct btfp_status btfp_status[ NBTFP ]; /* Per-instance state structure.*/

/****************************************************************************/
/* A dev_ops structure and identify routine are required for SunOS4.1.3
   and later.  This should not affect the ability of this code to be
   used with SunOS4.1.2 and earlier.
*/
int btfpidentify();                     /* Declare the identify routine.*/

struct  dev_ops btfp_ops = {            /* Required for SunOS4.1.3 & later.*/
        1,                              /* revision number              */
        btfpidentify,                   /* confirm device ID            */
        btfpattach,                     /* attach routine of driver     */
};
 
static int
btfpidentify(name)
        char *name;
{
        if( strcmp( name, "btfp" ) == 0)
        {   
            return( 1 );                /* Good match; device accepted. */
        }   
        else
        {   
            return( 0 );                /* No match; device rejected. */
        }   
}

/****************************************************************************/

/************************************************************************/
/*  btfpprobe() - This routine is called at system boot time.
                    To 'probe if the module is there', we read the bc63{57}'s
                    ID register.  If this read does not cause a bus error
                    and if its LSB value is different than 0x00, and also if
                    a write to the TIMEREQ register succeeds, then we assume a
                    bc63{57} is there and return the amount of VME memory the
                    bc63{57} uses, else we return 0 indicating device not there. 
*/
/************************************************************************/
btfpprobe( reg, unit )
    caddr_t reg;                    /* Address of the bc63{57} base register.*/
    int unit;                       /* The bc63{57} device number.          */
{
    register struct btfp_device *dp =
            (struct btfp_device *) reg; /* Init a pointer to device registers.*/    
    register short id_reg;              /* Places to read a module register.*/
    
 
    id_reg   = peek( (unsigned short *)&dp->btfp_id);   /* Read ID register.*/
    
    if( poke( (unsigned short *)&dp->btfp_timereq, 0 ) | /* Nothing at base.*/
        ( id_reg == -1 ) |              /* -1 indicates a bus error on read.*/
        ( id_reg & 0x00ff ) == 0x0000 ) /* ID reg 0 should be non-zero on init*/
    {
        printf( "btfpprobe(): btfp%d NOT found on the VME bus!\n", unit );
        return( 0 );                    /* Exit with an error indication.   */
    }
    else                                /* We assume we contacted a bc63{57}.*/
    {
        printf( "btfpprobe(): btfp%d is present on the VME bus.\n", unit );
        return( sizeof( struct btfp_device ) ); /* Return mem space it uses.*/
    }
} /* end of btfpprobe routine */


/*****************************************************************************/
/*	btfpattach() - This routine is called at system boot time.
			It initializes the bc63{57}'s interrupt vector and
			priority level, and sets the device status structure's
			interrupt indication to not waiting for an interrupt. 
			(The module comes up on system reset with interrupts 
			disabled.)  
*/	
/*****************************************************************************/
btfpattach( md )
	struct mb_device *md;		/* Receive a pointer to info struct.. */
					/* which was set up by config program.*/
{
					/* Init pointer to device structure. */
	register struct btfp_device *dp = ( struct btfp_device * ) md->md_addr;	
					/* Init pointer to module status struc*/
	register struct btfp_status *sp = &btfp_status[ md->md_unit ];
										
	if( md->md_intr )		/* Some vector was spec'd to config..*/
	{
		dp->btfp_vector = md->md_intr->v_vec; /* Load int. vector */
					/* Load the interrupt priority level */
		dp->btfp_level = (short)btfp_pritosplx( md->md_intpri);
		sp->intwait = 0;	/* Indicate not waiting for an int.*/
		sp->signalon  = 0;	/* Signal mechanism is turned off.*/
		sp->packeton  = 0;	/* Packet mode is turned off.	*/
		sp->debugon  = 0;	/* Debug mode is turned off.	*/
		sp->gps_synch  = 0;	/* Not synched with GPS packets. */
	}
	else				/* No interrupt vector in config file.*/
	{
		printf( 
		"btfpattach(): No interrupt vector in config file for btfp%d\n",
								md->md_unit );
	}
}

/*----------------------------------------------------------------------------*/
/*	I can't seem to get the configured interrupt priority level to 
	come from pritospl( md->mc_intpri ) so this forces level 3.  This is
	very likely where we want it, anyway.
*/
int
btfp_pritosplx( value )			/*Temporary to force int lvl.*/
	int value;
{
	return( BTFPINTLEVEL );		/* Force constant interrupt level.*/
					/* Defined in btfpu.h.*/
}

#endif sunos4
/*----------------------------------------------------------------------------*/
/* End of SunOS4 autoconfiguration 'wrapper' code.
*/
/************************************************************************/



/************************************************************************/
/************************************************************************/
/* Traditional UNIX entry points for ongoing driver operations...
	(tabstop=4 beyond this point)
*/
/************************************************************************/
/*                                                                     
  	btfpopen()- Called when an appication calls open(2).
  				If the incoming device minor number which is an index
  				to all of the btfpn's is larger than NBTFP which
  				was set by config, or if the btfpprobe routine did
  				not find this bc63{57}, then the open fails, 
				else it succeeds.
  			
  				The module's MASK register is written to 
  				disable external interrupts, the control register 0
				is set to disable interrupts, and the driver's
  				indicators for this device are initialized
  				accordingly.
*/                                                                      
/************************************************************************/

#ifdef sunos4
btfpopen( dev, flags )
	dev_t dev; 		 					/* Device (major/minor) number.	*/
	int flags;                      	/* Read/write flag.*/
{
	register int unit = minor( dev ); 	/* Get bc63{57} unit number.*/
	register struct mb_device *md = btfpinfo[ unit ];
										/* Pointer to our info strucrure.	*/
	register struct btfp_device *dp; 	/* Pointer to device structure.*/
	register struct btfp_status *sp; 	/* Pointer to module status structure.*/
#endif sunos4
#ifdef sunos5

static int
btfpopen( dev_t *devp, int flag, int otyp, cred_t *credp )
{
    struct btfp_device *dp;         	/* Pointer to device regs structure.*/
    struct btfp_status *sp;         	/* Pointer to module status struc. */
#endif sunos5

	unsigned short	ustemp, usmask;		/* For use turning off interrupts.*/
	int	s;								/* Interrupt level from spl().*/

	SHOW( "btfpopen:   Called!\n" );

#ifdef sunos4
	if( ( unit >= NBTFP ) || md == 0 || md->md_alive == 0 ) 
	{
		return( ENXIO );				/* Set no such device or address err.*/
	}
	dp = ( struct btfp_device * ) md->md_addr; /* Init ptr to device structure*/
	sp = &btfp_status[ md->md_unit ]; 	/* Init pointer to module status struc*/
#endif sunos4

#ifdef sunos5
    if( otyp != OTYP_CHR )              /* Support char device only.*/
            return( EINVAL );           /* Error return if not.*/

                                        /* Return error if attach was */
                                        /* not yet (successfully) run.*/
    if( ddi_get_soft_state( btfpstatep, getminor( *devp ) ) == NULL)
            return( ENXIO );

                                    	/* Get state/status structure pointer.*/
    sp = ddi_get_soft_state( btfpstatep, getminor( *devp ) );
    dp = sp->regp;                  	/* Working pointer to device regs.*/
#endif sunos5

										/* Do a successful, normal, open...	*/
	LOCKINT; 							/* Block interrupts over the following*/
	sp->intwait = 0;					/* Indicate not waiting for an int.*/
	sp->signalon  = 0;					/* Signal mechanism is turned off.*/
	sp->packeton  = 0;					/* Packet mode is turned off.*/
	sp->debugon  = 0;					/* Debug mode is turned off.*/

	dp->btfp_mask = (short)0;			/* Turn off all 5 interrupts.*/

	usmask = LOCKDIS | HBDIS | EVENTDIS | STRDIS; 
										/* Disable lockout, heartbeat */
										/* Event, and strobe interrupts */
	dp->btfp_cr0 = usmask;

#ifdef sunos4
 	sp->p_pgroup = (u.u_procp)->p_pid;  /* From the u struct (in case of kill)*/
#endif sunos4
	UNLOCKINT;							/* Turn interrupts back on.*/

	SHOW( "btfpopen:   Returning OK!\n" );
	return( 0 );						/* Else return open successful.*/

} /* end of btfpopen routine */


/************************************************************************/
/*                                                                      
  	btfpclose()- Called on the last close(2) system call to the device.
  				 This routine turns off interrupts from all 5 interrupt 
  				 sources on the bc63{57}, so none will occur when the
  				 device is not in use.  It resets the driver's status
  				 indicators accordingly, too.  Remember, close is only
  				 called on the LAST device to close of all that are
  				 open.
*/                                                                      
/************************************************************************/

#ifdef sunos4
btfpclose( dev, flags )
	dev_t dev;                          /* Major & minor device numbers.*/
	int flags;                          /* Read/write flag. 			*/
{
	register int unit = minor( dev );	/* Get bc63{57} unit number.		*/
	register struct mb_device *md = btfpinfo[ unit ];
										/* Pointer to our info strucrure.*/
	register struct btfp_device *dp = 
				( struct btfp_device * ) md->md_addr;	
										/* Init pointer to device structure.*/
    register struct btfp_status *sp = &btfp_status[ md->md_unit ];
                                    	/* Init pnter to module status struc. */
#endif sunos4
#ifdef sunos5

static int
btfpclose( dev_t dev, int flag, int otyp, cred_t *credp )
{
    struct btfp_device *dp;         	/* Pointer to device regs structure.*/
    struct btfp_status *sp;         	/* Pointer to module status struc. */
#endif sunos5

	unsigned short	ustemp, usmask;		/* For use turning off interrupts.*/
	int		s;							/* Interrupt level from spl().*/

#ifdef sunos5
    sp = ddi_get_soft_state( btfpstatep, getminor( dev ) );/* Status pointer*/
    dp = sp->regp;                      /* Working pointer to device regs.*/
#endif sunos5

	SHOW( "btfpclose:  Called!\n" );	/* Debug logging message.*/
	LOCKINT; 							/* Block interrupts over the following*/
	sp->intwait = 0;					/* Indicate not waiting for an int.*/
	sp->signalon  = 0;					/* Signal mechanism is turned off.	*/
	sp->packeton  = 0;					/* Packet mode is turned off.	*/
	sp->gps_synch  = 0;					/* Not synched with GPS packets. */
	dp->btfp_mask = 0;					/* Turn off all 5 interrupts.		*/
	usmask = LOCKDIS | HBDIS | EVENTDIS | STRDIS; /* Disable lockout+heartbeat*/
	dp->btfp_cr0 = usmask; 				/* Event, and strobe interrupts */
	sp->p_pgroup = 0;					/*No processes to signal on last close*/
	UNLOCKINT;							/* Turn interrupts back on.			*/
	return( 0 );						/* Return close successful.			*/

} /* end of btfpclose routine */

/************************************************************************/
/*
	send_packet - send a packet to the INFIFO.

		NOTE: packet must be null-terminated!!!!

		References MAX_SPIN to avoid infinite loop.

		NOTE: interrupts should be disabled in this routine. Since we
			  call it from btfpintr and elsewhere, we depend on the
			  caller to block the interrupts

		01/31/94..SGS..Pass in registers pointer and skip need for status
						structure pointer.  Hence simplifying code and
						making it portable.
*/
/************************************************************************/
int
send_packet( dp, packet )
struct btfp_device *dp;					/* Receive pointer to device regs.*/
unsigned char *packet;					/* Pointer to packet to send.*/
{
#ifdef OMIT_NOLONGER_USED
	register struct mb_device *md = btfpinfo[ unit ];
										/* Pointer to our info strucrure.*/
	register struct btfp_device *dp = 
				( struct btfp_device * ) md->md_addr;	
										/* Init pointer to device structure.*/
    register struct btfp_status *sp = &btfp_status[ md->md_unit ];
                                    	/* Init pnter to module status struc. */
#endif OMIT_NOLONGER_USED

	int loop_count;
	unsigned char *byte_ptr=packet;
	unsigned short ustemp;				/* For accessing ACK register */
	int		s;							/* Interrupt level from spl().	*/

	dp->btfp_ack = ACK_INFIFO;			/* Clear acknowledge bit */

	while (*byte_ptr)
	{
		dp->btfp_infifo = *byte_ptr;
		byte_ptr++;
	} /* end of while */

	dp->btfp_ack = ACK_INACT;			/* "Take-Action" bit */

										/* Wait for bc63{57} to take action */

										/* Note: no interrupt generated, */
										/* so we have to spin */
	loop_count = 0;
	while (loop_count <= MAX_SPIN)
	{
		ustemp = dp->btfp_ack;
		if ( ustemp & ACK_INFIFO ) /* Break if action taken */
		{
			break;
		}
		else
		{
			loop_count++;
		}
	}

	dp->btfp_ack = ACK_INFIFO;			/* Clear acknowledge bit */

	if (loop_count >= MAX_SPIN)
	{
		uprintf ("TIMEOUT (%d) waiting for INFIFO action\n", MAX_SPIN);
		return (1);						/* Return Error */
	}
#ifdef OMIT
	else
	{
		if (sp->debugon)
			uprintf ("%dLoop ", loop_count);
	}
#endif OMIT

	return (0);							/* Return No Problem */

} /* end of send_packet routine */


/************************************************************************/
/*                                                                      
  	btfpread()- Called when an application issues a read(2) call.
       

		PACKETOFF mode (default):

		This routine makes a time request.

  		Access the bc63{57}'s capture register to freeze the time
  		then return (converting BCD to ASCII) bytes of the time	
		to the caller, until the number of bytes he requested is
		reached or until we run	out of bytes to return.

		Sends 49 bytes maximum, separating each of the 16 bytes
		of time data with 3 spaces, and new-line and null 
		terminating the string.

      	This routine exits when 34 bytes have been sent in response
		to a read, regardless of what u.u_count is.  If it was not
		zero, then presumably the caller (e.g. cat()) will call
		again, but with u.u_offset = 33, which is our indication to
		just return without any transfer, which will result in the
		caller seeing 0 bytes transferred, which means EOF, by
		UNIX definition.
		
		For SunOS4/5 UNIX, iovec.iov_len (where struct iovec is pointed
		to in uio.uio_iov) replaces u.u_count of System VR3 UNIX, and 
		uio.uio_offset replaces u.u_offset.  ureadc() replaces passc().

		SIGNALON/SIGNALOFF modes have no effect on this mode.


		PACKETON mode/SIGNALOFF mode:

		Clear any pending interrupt status, and disable any interrupts.
		If a data packet is available, read the OUTFIFO into the caller's
		buffer and return it to the caller.  Otherwise, block until a 
		data packet is available from the OUTFIFO, then fill the buffer 
		supplied by the read call with the data from the OUTFIFO.  

		
		PACKETON mode/SIGNALON mode:

		If both PACKETON and SIGNALON modes are in effect, then
		the user must issue the ioctl calls to clear the
		DPA interrupt status (WINTSTAT) and enable DPA interrupts
		in the MASK register in a SIGNALON (non-blocking) WMASK 
		ioctl call.

		When the signal is received to indicate the DPA interrupt has
		occurred, the signal handler should issue the read(2) call,
		still leaving the driver in the SIGNALON mode. This will
		cause the btfpread() routine to immediately read the data
		packet from the OUTFIFO.

		/// (SunOS4 only)I do not understand how 'unit' can be out of range
			if it was checked on the open, but SUN driver manual
			says to check it here.
*/
/************************************************************************/

#define	MAXTBYTES	49				/* Size of returned time string.*/
#define	MAXABYTES	22				/* Size of converted time array.*/

#ifdef sunos4
btfpread( dev, uio )
	dev_t dev;						/* Major & minor device info. 	*/
	struct uio *uio;				/* Receive pointer to uio struc.*/
{
	register int unit = minor( dev );/* Get bc63{57} unit number.*/
	register struct mb_device *md = btfpinfo[ unit ];
									/* Pointer to our info strucrure.*/
	register struct btfp_device *dp = ( struct btfp_device * ) md->md_addr;	
									/* Init pointer to device structure.*/
    register struct btfp_status *sp = &btfp_status[ md->md_unit ];
                                   	/* Init pnter to module status struc. */
#endif sunos4

#ifdef sunos5
static int
btfpread(dev_t dev, struct uio *uio, cred_t *credp)
{
	struct btfp_device *dp; 		/* Pointer to device regs structure.*/
   	struct btfp_status *sp;         /* Pointer to module status struc. */
#endif sunos5

   	register struct iovec *iovp = uio->uio_iov;
   									/* Init pointer to iovec struct.*/
   	unsigned short dummy;			/* Dummy for use accessing captr.*/
   	unsigned short ustemp, ustemp2;	/* Temps for use accessing regs.*/
   	char timearray[ MAXABYTES ];	/* Place to put converted time.	*/
   	register i = 0;					/* Loop index thanks to K&R.	*/
    	register twi;				/* Time word index.*/
   	register tai;					/* Time array index.*/
   	unsigned char lochar;			/* Place to build the ASCII.*/
   	unsigned char hichar;			/* Ditto.*/
	int rtn;						/* Catcher for return code */
	int s;							/* return from spl()	*/
	int length;						/* Used for PACKETON read*/
	unsigned char *user_buffer;		/* pointer into user buffer */

   	union char_union 				/* For converting short to 'chars'.*/
   	{
   		unsigned char	two_chars[ 2 ];	/* Two characters holder.*/
   		unsigned short	time_word;	/* One of the time words.*/
   	} ch_union;
   	
	SHOW( "btfpread:   Called!\n" );/* Log message if debug is on.*/
#ifdef sunos5
									/* State/status structure pointer.*/
    sp = ddi_get_soft_state( btfpstatep, getminor( dev ) ); 
	dp = sp->regp;					/* Working pointer to device regs.*/
#endif sunos5

#ifdef sunos4
	if( unit >= NBTFP )				/* Beyond supportable range.*/
		return( ENXIO );			/* Return error indication.*/   	
#endif sunos4


	if (!sp->packeton)
	{
	/* PACKETOFF mode (default).  This means read time from module.*/

	dummy = dp->btfp_timereq;		/* Freeze the time via access to */
									/* the timereq register.*/
									/* Unpack bytes from btfp..	*/
   	timearray[ MAXABYTES - 1 ] = '\0';/* Set up zero termination.*/
   	timearray[ MAXABYTES - 2 ] = '\n';/* And new-line termination.	*/											

    tai = MAXABYTES - 3;            /* Fill time array back to front.*/
    for( twi = 4; twi >= 0; twi--)  /* Read time; BCD->ASCII.       */
    {
        ch_union.time_word = dp->btfp_time[ twi ];/* Get a time word.*/

        lochar = ch_union.two_chars[ 1 ];   /* Get two digits to process.   */
        hichar = ch_union.two_chars[ 1 ];   /* And into another var.        */
         
        lochar = ( lochar >> 4 ) | 0x30;    /* Convert a nibble to ASCII byte.*/        hichar = ( hichar & 0xf) | 0x30;    /* Convert the other nibble.    */
         
        timearray[ tai ] = hichar;          /* Move into the time array.    */
        tai--;                              /* Decrement the time array indx.*/
        timearray[ tai ] = lochar;          /* Move the other byte.         */
        tai--;                              /* Decrement the time array indx.*/
 
        lochar = ch_union.two_chars[ 0 ];   /* Get two digits to process.   */
        hichar = ch_union.two_chars[ 0 ];   /* And into another var.        */
         
        lochar = ( lochar >> 4 ) | 0x30;    /* Convert a nibble to ASCII byte.*/        hichar = ( hichar & 0xf) | 0x30;    /* Convert the other nibble.    */
         
        timearray[ tai ] = hichar;          /* Move into the time array.    */
        tai--;                              /* Decrement the time array indx.*/
        timearray[ tai ] = lochar;          /* Move the other byte.         */
        tai--;                              /* Decrement the time array indx.*/
        if( tai == 0 ) break;               /* End the loop after MAXTBYTES. */

    }
 

                                            /* Now send time string to caller.*/   
    i = 0;                                  /* Initialize byte index.       */
                            
    while ( iovp->iov_len )                 /* While application wants chars.*/
    {
        if( (i <= MAXABYTES - 1) &&         /* Deliver up to max bytes...   */
            (uio->uio_offset < MAXTBYTES) ) /* And stop on a later read.    */
        {
            if( i == 2 ||                   /* Between the time fields...   */
                i == 3 ||                   /* Day, hour, minute..etc...    */
                i == 6 ||                
                i == 8 ||
                i == 10 ||
                i == 12 ||
                i == 15 ||
                i == 18 ||
                i == 19)                    /* Insert spaces for readability.*/
            {
                ureadc( ' ', uio );         /* Send 3 spaces to the caller. */
                ureadc( ' ', uio );
                ureadc( ' ', uio );
            }
            ureadc( timearray[ i ], uio );  /* Move time byte to user space.*/
            i++;                            /* Move on to the next byte.    */
        }
        else                                /* We have no more to give...   */
        {                                   /* Or he doesn't want any more. */
            break;                          /* Exit the while and the read. */
        }
    }   
    return( 0 );                            /* Return no problem.*/

    }    /* End of PACKETOFF (read time) mode.*/
    else /* Start of PACKETON (means get packet from module) mode.*/
    {
                                            /* Beyond our packet buffer size.*/
        if( iovp->iov_len >= sizeof (sp->rpack_buffer) )
            return( ENXIO );                /* Return error indication.     */      
        if( !( sp->signalon ) )             /* Only block in SIGNALOFF mode */
        {
            LOCKINT; 						/* Block interrupts from our dev.*/
			ustemp = dp->btfp_intstat;		/* Read interrupt/status register.*/
			dp->btfp_intstat = ustemp;		/* Clear any interrupts.*/
            dp->btfp_mask = 0;   			/* Disable all bc635 interrupts.*/
            sp->intwait = 0;     			/* Tell btfpintr() NOT waiting.*/
			ustemp = dp->btfp_ack;			/* Get fifo status register.*/
			if( ustemp & ACK_DPA )			/* On entry here, had DPA already.*/
				goto readfifo;				/* So just read out the fifo.*/
            dp->btfp_mask = MASK_DPA;   	/* Enable DPA interrupt.*/
            sp->intwait = MASK_DPA;     	/* Tell btfpintr() we are waiting.*/

            while( sp->intwait )        	/* btfpintr() clears this on int.*/
            { 
                INTWAIT( sp->intwait ); 	/* Sleep/wait for DPA interrupt...*/
            }
											/* btfpintr() returns control here*/
  readfifo:									/* Read packet from board.*/
            sp->rpacki = 0;							/* Buffer index.*/
            ustemp = dp->btfp_ack;					/* Get fifo status.*/
            while( ( ustemp & ACK_MORE )         	/* Until FIFO is empty... */
         			&& ( sp->rpacki < MAXPACKET ) ) /* or running away */
            {                                   	/* get next byte.*/
                sp->rpack_buffer[ sp->rpacki++ ] = dp->btfp_outfifo;
                ustemp = dp->btfp_ack;          	/* Test for more data.*/
            } /* end of OUTFIFO read loop */
            dp->btfp_ack = ACK_DPA;         /* Clear possible DPA status.*/
            UNLOCKINT;                  	/* Allow our device intrpts again.*/

        } /* end of unique code to SIGNALOFF section, following code is common*/

		/* At this point, the interrupt routine or blocking section above 
			has read in the next packet (rpacki bytes) from the OUTFIFO 
			into the rpack_buffer, so just mode packet to caller buffer...
		*/

		if (sp->rpacki <= iovp->iov_len)	/* If fewer bytes than requested */
		{
			rtn = uiomove (sp->rpack_buffer,		/* Our buffer */
						 sp->rpacki,				/* # of bytes */
						 UIO_READ,					/* copy to user */
						 uio);						/* Give what we have */
			if (rtn != 0)
			{
				/* Error! */
				uprintf( "UIOMOVE (partial) error in btfpread\n" );
				return (rtn);
			}
		} /* end of "emptied our buffer first" case */
		else
		{
			/* user didn't give us a big enough buffer! */

			rtn = uiomove (sp->rpack_buffer,		/* Our buffer */
						 iovp->iov_len,				/* # of bytes */
						 UIO_READ,					/* copy to user */
						 uio);						/* Give what we have */
			if (rtn != 0)
			{
				/* Error! */
				uprintf ("UIOMOVE (full) error in btfpread\n");
				return (rtn);
			}

		} /* end of "filled user buffer" case */

		return( 0 );						/* Return no problem.*/
	} /* end of PACKETON mode */

} /* end of btfpread routine */


/************************************************************************/
/************************************************************************/
/*
	btfpwrite() - Called when an application does a write(2) call.


	PACKETOFF mode (default):

	This routine waits for the 1PPS pulse (unless in SIGNALON mode), 
	then loads time from user buffer.

	Accepts a buffer in UNIX's "date '+%j%H%M%S'" format, namely Julian date,
	hours, minutes, then seconds (with or without intervening spaces).

	For example, on the command line you would (as super user) type:
		# date '+%j%H%M%S' > /dev/btfp0

	Prepends this buffer with a 'B' (to indicate Load Major Time). 
	Note that this routine can accept the same format output by the 
	default read(2) call, because it strips off the ms, us, and ns fields.
		
			SIGNALON notes:

			If both PACKETOFF and SIGNALON modes are in effect, then
			the user must issue the ioctl calls to clear the
			1PPS interrupt status (WINTSTAT) and enable 1PPS interrupts
			in the MASK register in a SIGNALON (non-blocking) WMASK ioctl call.
			When the signal is received to indicate the 1PPS interrupt has
			occurred, the signal handler should issue the write(2) call,
			still leaving the driver in the SIGNALON mode. This will
			cause the btfpwrite() routine to immediately write the data
			packet (containing the date and time) to the INFIFO.


	PACKETON mode:

	Accepts a buffer from the user and transmits it directly to the INFIFO.

	SIGNALON/SIGNALOFF modes have no effect on this mode.

/************************************************************************/

#ifdef sunos4
btfpwrite(dev, uio)
	dev_t dev;							/* Major & minor device info. 	*/
	struct uio *uio;					/* Receive pointer to uio struc.*/
{
	register int unit = minor( dev );	/* Get bc63{57} unit number.*/
	register struct mb_device *md = btfpinfo[ unit ];
										/* Pointer to our info strucrure.*/
	register struct btfp_device *dp = ( struct btfp_device * ) md->md_addr;	
										/* Init pointer to device struct.*/
    register struct btfp_status *sp = &btfp_status[ md->md_unit ];
                                    	/* Init pnter to module status struc. */
#endif sunos4

#ifdef sunos5
static int
btfpwrite(dev_t dev, struct uio *uio, cred_t *credp)
{
    struct btfp_device *dp;         	/* Pointer to device regs structure.*/
    struct btfp_status *sp;         	/* Pointer to module status struc. */
#endif sunos5

   	register struct iovec *iovp = uio->uio_iov;
   											/* Init pointer to iovec struct.*/

	int rtn;							/* catcher for return code */
	int s;								/* spl() rtn code */
	unsigned char *user_buffer;			/* pointer into user buffer */
	register worki;						/* buffer index */
	int	num_bytes;						/* Number of user bytes */
	unsigned short usmask;				/* Used for register access */

    SHOW( "btfpwrite:  Called!\n" );	/* Log message if debug is on.*/
#ifdef sunos5
                                    	/* State/status structure pointer.*/
    sp = ddi_get_soft_state( btfpstatep, getminor( dev ) );
    dp = sp->regp;                  	/* Working pointer to device regs.*/
#endif sunos5
 
#ifdef sunos4
	if( unit >= NBTFP )					/* Beyond supportable range.*/
		return( ENXIO );				/* Return error indication.*/   	
#endif sunos4

	if (!sp->packeton)
	{
	/* PACKETOFF mode (default)(Means write time buffer to bc635) */

		num_bytes = iovp->iov_len;			/* Save number of user bytes */
											/* Do this now to save time later */
		rtn = uiomove (sp->work_buffer,		/* Our work buffer */
					 iovp->iov_len,			/* # of bytes */
					 UIO_WRITE,				/* copy from user */
					 uio);					/* Give what we have */
		if (rtn != 0)
		{
			/* Error ! */
			uprintf ("btfpwrite:  UIOMOVE (packetoff) Error!\n");
			return (rtn);
		}
											/* Leave room for 'B' */
		sp->wpack_buffer[0] = SOH;			/* Make null-terminated B packet */
		sp->wpack_buffer[1] = 'B';			
		sp->wpacki = 2;
		for (worki=0; worki <= num_bytes; worki++ )
											/* Eliminate white space, plus */
											/* ms, us, and ns fields */
		{
			if ( (sp->work_buffer[worki] != ' ')
				&& (sp->work_buffer[worki] != '\n')
				&& (sp->wpacki <= 10) ) 		/* (dddhhmmss takes 9 bytes) */
			{
				sp->wpack_buffer[sp->wpacki++] = sp->work_buffer[worki];
			}
		}
		sp->wpack_buffer[sp->wpacki++] = ETB;		/* Finish it off */
		sp->wpack_buffer[sp->wpacki] = (char) NULL;	/* with ETB and NULL */

											/* Wait for 1PPS to come around */
		if (! sp->signalon)					/* only block in SIGNALOFF mode */
		{

		LOCKINT; 							/*Block interrupts over following.*/
		if (sp->debugon)
			uprintf ("T");
		sp->intwait = MASK_1PPS;			/* Set waiting for a 1PPS int.*/
		dp->btfp_intstat = INTSTAT_1PPS;	/* Clear 1PPS int. status */
		usmask = dp->btfp_intstat;			/* Clear ALL int. status */
		dp->btfp_intstat = usmask;			/* ... */
		dp->btfp_mask = MASK_1PPS;			/* Set MASK reg. for 1PPS int.*/
		while ( sp->intwait )   		/* btfpintr() will clear this on int.*/
		{								/* Block, waiting for interrupt.*/
			INTWAIT( sp->intwait ); 
										/* Sleep till our interrupt occurs.*/
		}
		UNLOCKINT;				        /* Allow our device interrupts again.*/

		/* Interrupt routine should send the packet in wpackbuffer (?2/94.SGS)*/ 
		} /* end of (blocking) SIGNALOFF mode */


		LOCKINT; 						/*Block interrupts over the following.*/
		if ( (send_packet (dp, sp->wpack_buffer)) == 1)/* Load time buffer */
		{
			UNLOCKINT;				    /* Allow our device interrupts again.*/
			uprintf ("btfpwrite:  Error from send_packet of wpack_buffer\n");
			return (EIO);
		}
		UNLOCKINT;				        /* Allow our device interrupts again.*/


		return( 0 );					/* Return no problem.*/
	}
	else
	{
	/* PACKETON mode (Means send general user packet to bc635)*/

		if( iovp->iov_len >= sizeof (sp->wpack_buffer) ) /* > packet buf size.*/
			return( ENXIO );			/* Return error indication.		*/   	
		num_bytes = iovp->iov_len;		/* Save number of user bytes */

		rtn = uiomove (sp->wpack_buffer,/* Our buffer */
					 iovp->iov_len,		/* # of bytes */
					 UIO_WRITE,			/* copy from user */
					 uio);				/* Give what we have */
		if (rtn != 0)
		{
			/* Error! */
			uprintf ("btfpwrite:  UIOMOVE (packeton) error!\n");
			return (EIO);
		}

										/* Make into null-terminated packet */
		sp->wpack_buffer[num_bytes] = (unsigned char) NULL;	

		LOCKINT; 						/* Block interrupts .*/

		if ( (send_packet (dp, sp->wpack_buffer)) == 1) /* Send raw packet */
		{
			UNLOCKINT;					/* Turn interrupts back on.	*/
			return (EIO);
		}
		UNLOCKINT;						/* Turn interrupts back on.*/

		return( 0 );					/* Return no problem.*/
	} /* end of PACKETON mode */

} /* end of btfpwrite routine */


/************************************************************************/
/*                                                                      
  	btfpioctl()- Called when an application does an ioctl(2) call.
  								                                   
  					Generally, the bc63{57}'s registers are written or read
  					by these routines.
					The SunOS4 ioctl(2) call provides the _IOx macros which
					tell it whether the arg (data) item is an int or a 
					pointer and whether the data should be copied into
					or out of the user space or both, and the length of
					the data (number of bytes).  Copyin and copyout are
					not needed since the OS does this for us.  In SunOS5,
					however, the opposite is true! (see macros in btfpd.h).
					
					NOTE:  If we are in SIGNALON mode, we do not block when
					an external interrupt enable (to MASK) is received,
					else we do block.

                    NOTE:  We are enabling the interrupt after the cause
                    of it may have happened, e.g. strobe may already
                    have gone off.  Hence an application could sleep
                    forever.

	01/28/94..SGS..With SunOS5 ioctl's, copyin/copyout must be used
					to move data to and from user space, since 
					the OS no longer does this associated with 'arg'
					(when arg is a pointer...if arg is passed as an
					int directly on the call, copyin is NOT required).
					Since we may be on a VME bus adaptor, such 
					as the Solflower SFVME, and since copyin/copyout
					do not always do half-word transfers, (if
					they think they are talking to the SBus)
					we must intermediate-buffer all data xfers (where
					'arg' is a pointer) to and from the bc635 registers.
*/                                                                   
/************************************************************************/
#ifdef sunos4
btfpioctl( dev, cmd, arg, flag )
	dev_t dev;                      /* Major & minor device info.*/
	int cmd;                        /* The command.*/
	caddr_t arg;                   	/* The data argument or pointer to it.*/
	int	flag;						/* Unused in this driver.*/
{
	register int unit = minor( dev ); /* Get bc63{57} unit number.*/
	register struct mb_device *md = btfpinfo[ unit ];
									/* Pointer to our info strucrure.*/
	register struct btfp_device *dp = 
			( struct btfp_device * )md->md_addr;	
									/* Init pointer to device structure.*/
	register struct btfp_status *sp = &btfp_status[ md->md_unit ];
									/* Init pointer to module status struc.	*/
	register struct btfp_device *up = (struct btfp_device *)arg;
									/* READALL receives such a pointer.*/
	register struct btfp_time *tp = (struct btfp_time *)arg;
									/* Pointer to time register structure.*/
	register struct btfp_event *ep = (struct btfp_event *)arg;
									/* Pointer to event register structure.*/
	register struct btfp_strobe *strp = (struct btfp_strobe *)arg;
									/* Pointer to strobe register structure.*/
#endif sunos4

#ifdef sunos5
static int 
btfpioctl( dev_t dev,int cmd, int arg, int flag, cred_t *credp, int *rvalp )
{
	struct btfp_device *dp;         /* Pointer to device regs structure.*/
	register struct btfp_status *sp;/* Pointer to module status struc. */

	struct btfp_device bdev;		/* Local device reg struct for copies.*/
	struct btfp_time   btime;		/* Local time struct for copies.*/
	struct btfp_event  bevent;		/* Local event struct for copies.*/
	struct btfp_strobe bstrobe;		/* Local strobe structure for copies.*/
	register struct btfp_device *up = &bdev; /* Pointer to device regs struct.*/
	register struct btfp_time *tp =  &btime; /* Pointer to time reg struct.*/
	register struct btfp_event *ep = &bevent;/* Pointer to event reg struct.*/
	register struct btfp_strobe *strp = &bstrobe; /* Ptr to strobe reg struc.*/
#endif sunos5

	int		s;						/* Interrupt level from spl().*/
	register int maskin;			/* Interrupt mask received from caller.*/
   	unsigned short ustemp, ustemp2;	/* Temps for use accessing regs.*/
   	unsigned int uitemp;			/* Temp for use in returning values.*/
	unsigned short 	usmask = 
			MASK_INT0 | MASK_INT1 | MASK_INT2 | MASK_INT3 | MASK_INT4;
									/* Mask for turning off device interrupts.*/ 
#ifdef OMIT
	SHOW( "btfpioctl:  Called!\n" );/* Message to console if debug is on.*/
#endif OMIT
#ifdef sunos5
                                    /* State/status structure pointer.*/
	sp = ddi_get_soft_state( btfpstatep, getminor( dev ) );
	dp = sp->regp;              	/* Working pointer to device regs.*/
#endif sunos5
 

	switch (cmd) {					/* Handle specific incoming commands.*/
	
	case RIDR:						/* Read the ID register.*/
   		ustemp = dp->btfp_id;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;
	
	case RDTR:						/* Read the Device Type register.*/
   		ustemp = dp->btfp_device;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;
	
	case RSR:						/* Read the Status register.*/
   		ustemp = dp->btfp_status;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case WCR:						/* Write to the Control register.*/
		INTIN( arg, dp->btfp_control ); /* Directly from the user's argument.*/
		break;
	
	case RTIMEREQ:					/* Read the Time request register.*/
   		ustemp = dp->btfp_timereq;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case RTIME0:					/* Read time word zero.*/
		ustemp = dp->btfp_time[ 0 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case RTIME1:					/* Read time word one.*/
		ustemp = dp->btfp_time[ 1 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case RTIME2:					/* Read time word two.*/
		ustemp = dp->btfp_time[ 2 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case RTIME3:					/* Read time word three.*/
		ustemp = dp->btfp_time[ 3 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case RTIME4:					/* Read time word four.*/
		ustemp = dp->btfp_time[ 4 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case REVENT0:					/* Read event word zero.*/
		ustemp = dp->btfp_event[ 0 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case REVENT1:					/* Read event word one.*/
		ustemp = dp->btfp_event[ 1 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case REVENT2:					/* Read event word two.*/
		ustemp = dp->btfp_event[ 2 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case REVENT3:					/* Read event word three.*/
		ustemp = dp->btfp_event[ 3 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case REVENT4:					/* Read event word four.*/
		ustemp = dp->btfp_event[ 4 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case WSTROBE1:					/* Write strobe word one.*/
   		INTIN( arg, dp->btfp_strobe1 );/* Directly from the user's argument.*/
		break;

	case WSTROBE2:					/* Write strobe word two.*/
   		INTIN( arg, dp->btfp_strobe2 );/* Directly from the user's argument.*/
		break;

	case WSTROBE3:					/* Write strobe word three.*/
   		INTIN( arg, dp->btfp_strobe3 );/* Directly from the user's argument.*/
		break;

	case RUNLOCK:					/* Read Unlock register.*/
   		ustemp = dp->btfp_unlock;	/* Contents of the register matter not*/
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;
	
	case RACK:						/* Read the ACK register.*/
   		ustemp = dp->btfp_ack;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only */
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case WACK:						/* Write to the ACK register.*/
   		INTIN( arg, dp->btfp_ack ); /* Directly from the user's argument.*/
		break;

	case WCR0:						/* Write control register zero.*/
		INTIN( arg, dp->btfp_cr0 ); /* Directly from the user's argument.*/
		break;

	case RCR0:						/* Read control register zero.*/
		ustemp = dp->btfp_cr0;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;
	
	case ROUTFIFO:					/* Read the OUTFIFO register.*/
   		ustemp = dp->btfp_outfifo;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case WINFIFO:					/* Write to the INFIFO register.*/
   		INTIN( arg, dp->btfp_infifo );/* Directly from the user's argument.*/
		break;
	
	case RMASK:						/* Read the MASK register.*/
   		ustemp = dp->btfp_mask;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only	*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case WMASK:						/* Write to the MASK register.*/
		INTIN( arg, maskin );		/* Get caller's mask field.*/
		if( !( maskin & (int)usmask ) ) /* No interrupt enable requested.*/
		{
									/* Set device mask register...*/
			INTIN( arg, dp->btfp_mask ); /* Directly from the user's argument.*/
		}
		else if( !sp->signalon )	/* Signal sending is NOT enabled,*/
									/* but an interrupt wait is requested.*/
		{
			LOCKINT; 				/*Block interrupts over the following.*/
										
			if( ( usmask & MASK_DPA )				/* We have DPA already */
				&& (dp->btfp_mask & (unsigned short)maskin))/*& user wants DPA*/
			{
				if( !( sp->gps_synch ) )			/* Are we in synch w/GPS?*/
				{
					dp->btfp_ack = ACK_CLEAR;		/* Clear OUTFIFO.*/
													/* When we get the next */
													/* DPA, the interrupt*/
													/* routine will get us*/
													/* synch-ed up with GPS*/
													/* and set the flag*/
				}
				dp->btfp_ack = ACK_DPA;				/* Clear DPA status */
			} /* end of "We have DPA-already" section */

			sp->intwait = maskin;			/* Indicate waiting for an int.*/
			dp->btfp_intstat = maskin;		/* Clear any existing interrupt*/
									/* Save the interrupts we're waiting for*/
			usmask = dp->btfp_intstat;				/* Clear ALL int. status*/
			dp->btfp_intstat = usmask;				/* ... */
			dp->btfp_mask = maskin; /* Enable bc63{57} interrupt(s).*/

			while( sp->intwait )    /* btfpintr() will clear this on int.*/
			{						/* Block, waiting for bc63{57} interrupt.*/
				INTWAIT( sp->intwait ); 
									/* Sleep till our interrupt occurs.*/
			}
			UNLOCKINT;				/* Allow our device interrupts again.*/
		}
		else						/* Signal sending IS enabled.*/
		{
			LOCKINT; 				/*Block interrupts over the following.*/
			if( usmask & MASK_DPA )					/* DPA already? */
			{
				if( !(sp->gps_synch) )				/* Are we in synch w/GPS?*/
				{
					dp->btfp_ack = ACK_CLEAR;		/*  Clear OUTFIFO*/
													/* When we get the next */
													/* DPA, the interrupt   */
													/* routine will get us  */
													/* synch-ed up with GPS */
													/* and set the flag     */
				}
				dp->btfp_ack = ACK_DPA;				/* Clear DPA status*/

			} /* end of "DPA-already" section */

			sp->intwait = maskin;				/* We will not be blocking.*/
			dp->btfp_intstat = maskin;			/* Clear an existing interrupt*/
			usmask = dp->btfp_intstat;			/* Clear ALL int. status */
			dp->btfp_intstat = usmask;			/* ... */
			dp->btfp_mask = maskin; /* Enable bc63{57} interrupt(s) etc.*/
			UNLOCKINT;				/* Allow interrupts from our device again.*/
									/* Return without blocking with a sleep.*/
		}
		break;
	
	case RINTSTAT:					/* Read the INTSTAT register.*/
   		ustemp = dp->btfp_intstat;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only */
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case WINTSTAT:					/* Write to the INTSTAT register.*/
   		INTIN( arg, dp->btfp_intstat );/* Directly from the user's argument.*/
		break;
	
	case RVECTOR:					/* Read the VECTOR register.*/
   		ustemp = dp->btfp_vector;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only */
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;
	
	case RLEVEL:					/* Read the LEVEL register.*/
   		ustemp = dp->btfp_level;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only*/
		INTOUT( uitemp, arg );		/* Move it into the return argument.*/
		break;

	case READTIME:					/* Capture and read all time words.	*/
		ustemp		 		= dp->btfp_timereq;		/* Capture time.*/
		tp->btfp_time[ 0 ] 	= dp->btfp_time[ 0 ];	/* Time word 0.	*/
		tp->btfp_time[ 1 ] 	= dp->btfp_time[ 1 ];	/* Time word 1.	*/
		tp->btfp_time[ 2 ] 	= dp->btfp_time[ 2 ];	/* Time word 2.	*/
		tp->btfp_time[ 3 ] 	= dp->btfp_time[ 3 ];	/* Time word 3.	*/
		tp->btfp_time[ 4 ] 	= dp->btfp_time[ 4 ];	/* Time word 4.	*/
		COPYOUT( tp, arg, struct btfp_time ); 		/* Copy to caller.*/
		break;

	case READEVENT:					/* Read all event words.*/
		ep->btfp_event[ 0 ] = dp->btfp_event[ 0 ];	/* Event word 0.*/
		ep->btfp_event[ 1 ] = dp->btfp_event[ 1 ];	/* Event word 1.*/
		ep->btfp_event[ 2 ] = dp->btfp_event[ 2 ];	/* Event word 2.*/
		ep->btfp_event[ 3 ] = dp->btfp_event[ 3 ];	/* Event word 3.*/
		ep->btfp_event[ 4 ] = dp->btfp_event[ 4 ];	/* Event word 4.*/
		COPYOUT( ep, arg, struct btfp_event );		/* Copy to caller.*/
		break;

	case WRITESTROBE:				/* Write to the Strobe registers.*/
		dp->btfp_cr0 = STRDIS;		/* First disable strobe output*/
		COPYIN( arg, strp, struct btfp_strobe );	/* Get words from caller.*/
		dp->btfp_strobe1 = strp->btfp_strobe[0];	/* Strobe word 1*/
		dp->btfp_strobe2 = strp->btfp_strobe[1];	/* Strobe word 2*/
		dp->btfp_strobe3 = strp->btfp_strobe[2];	/* Strobe word 3*/
		break;

	case READALLREGS:				/* Read all bc63{57} regs.*/
		up->btfp_id		= dp->btfp_id;			/* ID register.		*/
		up->btfp_device = dp->btfp_device;		/* Device Type register.	*/
		up->btfp_status	= dp->btfp_status;		/* Status register.		*/

		up->btfp_timereq	= dp->btfp_timereq;		/* Capture the time.	*/
		up->btfp_time[ 0 ] 	= dp->btfp_time[ 0 ];	/* Time word 0.			*/
		up->btfp_time[ 1 ] 	= dp->btfp_time[ 1 ];	/* Time word 1.			*/
		up->btfp_time[ 2 ] 	= dp->btfp_time[ 2 ];	/* Time word 2.			*/
		up->btfp_time[ 3 ] 	= dp->btfp_time[ 3 ];	/* Time word 3.			*/
		up->btfp_time[ 4 ] 	= dp->btfp_time[ 4 ];	/* Time word 4.			*/

		up->btfp_event[ 0 ] = dp->btfp_event[ 0 ];	/* Event word 0.		*/
		up->btfp_event[ 1 ] = dp->btfp_event[ 1 ];	/* Event word 1.		*/
		up->btfp_event[ 2 ] = dp->btfp_event[ 2 ];	/* Event word 2.		*/
		up->btfp_event[ 3 ] = dp->btfp_event[ 3 ];	/* Event word 3.		*/
		up->btfp_event[ 4 ] = dp->btfp_event[ 4 ];	/* Event word 4.		*/

		up->btfp_unlock 	= dp->btfp_unlock;		/* UNLOCK register.	*/
	
		up->btfp_ack 		= dp->btfp_ack;			/* ACK register.	*/
		up->btfp_cr0 		= dp->btfp_cr0;			/* Control register 0.	*/
		up->btfp_outfifo	= dp->btfp_outfifo;		/* OUTFIFO register.	*/
		up->btfp_mask		= dp->btfp_mask;		/* MASK register.	*/
		up->btfp_intstat	= dp->btfp_intstat;		/* INTSTAT register.	*/
		up->btfp_vector		= dp->btfp_vector;		/* VECTOR register.	*/
		up->btfp_level		= dp->btfp_level;		/* LEVEL register.	*/
		COPYOUT( up, arg, struct btfp_device );		/* Copy to caller.*/
		break;

#ifdef SIGNALS
	case SIGNALON:					/* Turn on signal sending on interrupts.*/
		sp->signalon = 1;			/* Indicate we should send signals on ints*/
		break;
#endif SIGNALS
		
	case SIGNALOFF:					/* Turn off signal sending on interrupts.*/
		LOCKINT; 					/* Block device interrupts*/
		sp->signalon = 0;			/* Stop sending signals on interrupts.	*/
		sp->intwait = 0;			/* Indicate we are not awaiting any int.*/
		dp->btfp_mask = 0;			/* Turn off all 5 interrupts.	*/
		UNLOCKINT;					/* Allow interrupts from our device again.*/
		break;

	case PACKETON:					/* Turn on packet mode for read/write.*/
		sp->packeton = 1;			
		break;

	case PACKETOFF:					/* Turn off packet mode for read/write.*/
		sp->packeton = 0;			
		break;
		
	case DEBUGON:
		sp->debugon = 1;			/* Enable driver debugging */
		break;
		
	case DEBUGOFF:
		sp->debugon = 0;			/* Disable driver debugging */
		break;

	default:							
		return( ENOTTY );			/* Return this is not a tty type device.*/
		break;
	}
	return( 0 );					/* Must return OK if no error found.	*/

} /* end of btfpioctl routine */


/************************************************************************/
/*                                                 
  	btfpintr()- Called when the bc63{57} interrupts the processor.
  								                           
  					This routine assumes that the interrupt level
  					on entry is the same as for the device, so that
  					we do not have to do an spl call.  If 
  					btfp_status[ unit].intwait does not match the interrupt
					which has just occurred, we were not
  					expecting this interrupt and we print an error message
  					and return after masking off module interrupt bits.  
  					Otherwise, we clear btfp_status[ unit].intwait,
  					clear the board interrupt bits, and wakeup the sleeping
  					ioctl WMASK routine, unless SIGNALON mode is in effect,
  					in which case we read in a data packet if applicable
					and/or generate a signal and exit.
*/                                                                
/************************************************************************/
#ifdef sunos4
btfpintr( unit ) 
	int unit;                       /* bc63{57} unit number.*/
{
	register struct mb_device *md = btfpinfo[ unit ];
									/* Pointer to our info strucrure.*/
	register struct btfp_device *dp = ( struct btfp_device * ) md->md_addr;	
									/* Init pointer to device structure.*/
	register struct btfp_status *sp = &btfp_status[ md->md_unit ];
									/* Init pointer to module status struc.*/
	register int s;					/* Interrupt level holder.*/
#endif sunos4

#ifdef sunos5
static unsigned int
btfpintr( caddr_t unit )			/* ddi_add_intr specs unit as argument.*/
{
    struct btfp_device *dp;         /* Pointer to device regs structure.*/
    struct btfp_status *sp;         /* Pointer to module status struc. */
	procset_t ps;					/* ///Used to send signal via sigsendset()*/
	sigsend_t ss;					/* ///Used to send signal via sigsendset()*/
#endif sunos5

	unsigned short 	ustemp, ustemp2; /* Temps for register access */
	unsigned short 	usmask = 
					MASK_INT0 | MASK_INT1 |MASK_INT2 | MASK_INT3 | MASK_INT4;
									/* Mask of module interrupt enable bits.*/

    SHOW( "btfpintr:   Called!\n" );/* Log message if debug is on.*/
#ifdef sunos5
                                    /* State/status structure pointer.*/
    sp = ddi_get_soft_state( btfpstatep, (unsigned int)unit );
    dp = sp->regp;                  /* Working pointer to device regs.*/
#endif sunos5

	LOCKINT;								/* Block interruption in here.*/

	ustemp = dp->btfp_intstat;				/* Read INTSTAT register.*/

#ifdef sunos5
	if( !( usmask && ustemp ) )				/* Did our device interrupt?*/
	{
		UNLOCKINT;							/* If not, this might be a poll...*/
		return( DDI_INTR_UNCLAIMED );		/* Return interrupt not ours.*/
	}
#endif sunos5
	
	if( !(sp->intwait & ustemp) )			/* Were not waiting for this int!*/
	{
		dp->btfp_mask = 0;				 	/* Turn off all 5 interrupts.*/
		dp->btfp_intstat =  usmask; 		/* Turn off all 5 interrupts.*/
		uprintf( "btfpintr:   Unexpected interrupt from btfp%d\n", (int)unit );
		UNLOCKINT;							/* Release interrupt lock.*/
#ifdef sunos4
		return(0);							/* Return to caller.*/
#endif sunos4
#ifdef sunos5
		return( DDI_INTR_CLAIMED );			/* But must 'claim' intr anyhow.*/
#endif sunos5
	}
	else if( !sp->signalon )				/* We were waiting for an int..	*/
	{										/* and we were blocked doing so.*/
		sp->intwait = 0;					/* Indicate we received an interr.*/
		if (ustemp & INTSTAT_DPA)
		{
			/* For DPA interrupt, remember to clear the ACK_DPA bit */
			dp->btfp_ack = ACK_DPA;
		}

		dp->btfp_mask = 0;					/* Turn off all 5 interrupts.*/
		dp->btfp_intstat = usmask; 			/* Turn off any interrupts.	*/
		WAITWAKEUP( sp->intwait );			/* Allow the upper-half to proceed*/
	}
	else									/* Waiting for an interrupt but in*/
	{										/* SIGNALON mode...	*/
		dp->btfp_intstat = usmask; 			/* Turn off any interrupts.	*/

        /* If DPA is here... */
        if (ustemp & INTSTAT_DPA)
        {
 
            if (sp->debugon)
            {
                if (ustemp & INTSTAT_DPA)
                    uprintf ("d");
            }
 
            /* ...AND the user wanted to look at it */
            if (sp->intwait & MASK_DPA)
            {
                if (sp->debugon)
                {  
                    if (ustemp & INTSTAT_DPA)
                        uprintf ("D");
                }
 
                dp->btfp_ack = ACK_DPA;             /*Clear DPA bit in ACK reg*/ 
                /* NOTE: SIGNALON mode has already waited for the DPA signal, */                /* thus its read just reads the available packet */
 
                /* Now fill packet buffer from OUTFIFO */
 
                sp->rpacki = 0;
 
                /* We must read the first byte to activate the ACK_MORE flag */
                /* and to see if we are out of synch with the TANS packets.  */
                /* If the first two bytes don't look like the start of a     */                 /* packet, just clear the OUTFIFO again.                     */  
                if (!sp->gps_synch)
                {  
													/* Get first byte */
					sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;	
					if( sp->rpack_buffer[sp->rpacki - 1] != (unsigned char)DLE )
					{
						/* Bogus packet...clear OUTFIFO */
						dp->btfp_ack = ACK_CLEAR;		/* Clear OUTFIFO */
					}
					else		
					{
					/* First byte OK...look at next byte: is it a packet ID? */
						sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;	
						if( ( sp->rpack_buffer[sp->rpacki - 1] == 
														(unsigned char) DLE )
							|| ( sp->rpack_buffer[sp->rpacki - 1] == 
														(unsigned char) ETX ) )
						{
						/* Bogus packet...clear OUTFIFO */
							dp->btfp_ack = ACK_CLEAR;	/* Clear OUTFIFO */
						}
						/* else packet is OK...read it in */
				
					} /* end of first byte OK section */

					if (sp->debugon)
						uprintf ("S");
					sp->gps_synch = 1;					/*We're synched up now*/

				} /* end of "need to synch up with the GPS" section */

				/* NOTE: ACK_MORE bit on means "more bytes in OUTFIFO" */
				ustemp = dp->btfp_ack;
				while ( (ustemp & ACK_MORE) 		/* Until FIFO is empty... */
					&& (sp->rpacki < MAXPACKET) )	/* or running away */
				{									/* Get next byte.*/
					sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;	
					ustemp = dp->btfp_ack;			/* Any more? */
			
				} /* end of OUTFIFO read loop */
		
				if (sp->rpacki == 0)
				{
					if (sp->debugon)
						uprintf ("0");
													/* Wait for nxt DPA intrpt*/
					UNLOCKINT;						/* Release interrupt lock.*/
#ifdef sunos4
					return;
#endif sunos4
#ifdef sunos5
					return( DDI_INTR_CLAIMED );
#endif sunos5
				}

			} /* end of "user wanted to see DPA" section */
			else
			{
				ustemp &= ~INTSTAT_DPA;			/* Other interrupts happen?*/
				if ( !(ustemp & sp->intwait) )	/* If not our interest, exit.*/
				{
					UNLOCKINT;					/* Release interrupt lock.*/
#ifdef sunos4
					return;
#endif sunos4
#ifdef sunos5
					return( DDI_INTR_CLAIMED );
#endif sunos5
				}
			/* else, fall through to send the user a signal */
			}

		} /* end of DPA section */

		/* if no DPA, we must have matched on another INTSTAT bit */

#ifdef sunos4
		gsignal( sp->p_pgroup, SIGUSR1 );	/* Send signal to the application.*/
#endif sunos4
#ifdef sunos5
/* The following code does not work.  As of now, do not know how to
	generate a signal to the caller from here!
*/
		ps.p_op = POP_OR;					/* Fill struct with proc ID info.*/
        ps.p_lidtype = sp->p_pgroup;
        ps.p_lid = P_PID;
        ps.p_lid = P_ALL;
        ps.p_ridtype = sp->p_pgroup;
        ps.p_rid = P_PID;
        ps.p_rid = P_ALL; /*///*/
		ss.sig = SIGUSR1;					/* Fill in signal type structure.*/
		ss.perm = 0;						/* No real idea WHAT these are!!!/
		ss.checkperm = 0;
		ss.sicode = 0;
		ss.value = 0;
		sigsendset( &ps, &ss ); 			/* Send signal to application.*/
											/* /// How does this work?????*/
#endif sunos5
	} /* end of SIGNALON section */
	UNLOCKINT;								/* Release interrupt lock.*/
#ifdef sunos4
	return;
#endif sunos4
#ifdef sunos5
	return( DDI_INTR_CLAIMED );
#endif sunos5

} /* end of btfpintr routine */


/*****************************************************************************/
/*****************************************************************************/
