/****************************************************************************/
/*
	The include file for the bc635VME/bc350VXI SunOS-UNIX device driver.
	The device name is 'btfp' for 'BANCOMM time and frequency processor'.

	Copyright (C) 1994, BANCOMM, San Jose, CA.  All rights reserved.
                                                                      
  	Author:	Steve Schoggen

		S. G. Schoggen & Company
		P. O. Box 246
		Little Silver, NJ  07739-0246	   908-530-4659

		Winter, 1994
*/
/****************************************************************************/

/*	Required system include files...
*/
#ifdef sunos4
#include <sys/param.h>              /* System typedefs, PRIO defs, etc.     */
#include <sys/file.h>
#include <sys/dir.h>
#include <sys/user.h>               /* Current process data ('u' struct).   */
#include <sys/uio.h>                /* The uio structure definition.        */
#include <sys/signal.h>             /* Signal definitions.                  */
#include <machine/psl.h>
#include <sundev/mbvar.h>
#include "errno.h"                  /* Standard error returns- EINVAL, etc. */
#include "btfp.h"                   /* Generated by 'config'- has Nbtfp.    */
#endif sunos4

#ifdef sunos5
#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/signal.h>
#include <sys/uio.h>
#include <sys/buf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/kmem.h>
#include <sys/poll.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/stat.h>
#include <sys/procset.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>
#endif sunos5

#ifdef sunos4
#define SIGNALS			/* Only sunos4 supports signals as of now.///*/	
#endif sunos4


									
/****************************************************************************/
/*      bc63{57} device 'soft' state information structure.
*/
struct btfp_status
{
	int 	intwait;        /* Copy of MASK interrupt enable bits.*/
	int     signalon;       /* True if signal sending is enabled.*/
        int     packeton;       /* True if packet mode is enabled.*/
        int     debugon;        /* True if debug mode is enabled.*/
        short p_pgroup;         /* User process group leader(u.u_proc)*/
        int gps_synch;          /* True if synched with GPS packets */
        int rpacki;             /* Read Packet buffer index.*/
        int wpacki;             /* Write Packet buffer index.*/
        unsigned char rpack_buffer[MAXPACKET+1]; /* used for read calls*/
        unsigned char wpack_buffer[MAXPACKET+1]; /* for write calls.*/
        unsigned char work_buffer [MAXPACKET+1]; /* used to build pckt.*/
#ifdef sunos5
        dev_info_t *dip;        /* Device info handle (for ddi/ddk stuff).*/
	ddi_iblock_cookie_t  iblock_cookie; /* For mutexes- "interrupt block".*/
	ddi_idevice_cookie_t idevice_cookie;/* Device interrupt level.*/
        kmutex_t                mutex;	/* Mutual exclusion lock (splx).*/
        kcondvar_t              cv;	/* Condition variable (sleep/wakeup).*/
#endif sunos5
	struct btfp_device *regp;/* Pointer to device registers structure.*/
};
#ifdef sunos5
struct interruptinfo 			/* Contents of idevice_cookie...*/
{
	unsigned short vector;		/* The VME vector number.*/
	unsigned short priority;	/* The interrupt priority level.*/
};
#endif sunos5

/*****************************************************************************/
/* Macros to clarify portability between SunOS4 and SunOS5...
*/
#define DDI_CHARACTER DDI_PSEUDO	/* Do not understand ddi_create_minor!*/

/*---------------------------------------------------------------------------*/
/* Macro to reduce debug logging to a single, small line...
*/
#ifdef sunos4
#define SHOW( m ) \
		if( btfpdebug ) \
			printf( m )
#endif sunos4

#ifdef sunos5 
#define SHOW( m ) \
                if( btfpdebug ) \
                        cmn_err( CE_CONT, m )
#endif sunos5

/*---------------------------------------------------------------------------*/
/* For ioctl routine, moving an int to the user space from the device...
*/
#ifdef sunos4
#define INTOUT( i, a ) \
	*(int *)a = i
#endif sunos4

#ifdef sunos5
#define INTOUT( i, a ) \
	copyout( (caddr_t)&i, (caddr_t)a, sizeof( int ) )
#endif sunos5	


/* For ioctl routine, moving an int from the user space to device...
*/
#ifdef sunos4
#define INTIN( a, i ) \
	i = *(int *)a
#endif sunos4

/* For sunos5 we must ensure a half-word operation on the bus(es).
	This may not be needed///test using arg directly.

	In any case NOTE-NOTE-NOTE:  The ioctl 'arg' is an int if
	an int is passed in the call, otherwise it is a pointer.
	If it is an int, it need not be copyin'd; it may be used
	as is.  Only if it is a pointer is copyin (and copyout) needed!!!	
					
*/
#ifdef sunos5
#ifdef OMIT
#define INTIN( a, i ) \
        ustemp = (unsigned short)a \
        i = ustemp
#endif OMIT
#define INTIN( a, i ) \
        i = (unsigned short)a
#endif sunos5

/*---------------------------------------------------------------------------*/
/* For ioctl routine...general copyin/copyout macros which are no-op on sunos4.
*/
#ifdef sunos4
#define COPYOUT( xx, yy, zz ) ;
#endif sunos4

#ifdef sunos5
#define COPYOUT( from, to, thing ) \
	copyout( (caddr_t)from, (caddr_t)to, sizeof( thing ) )
#endif sunos5

#ifdef sunos4
#define COPYIN( xx, yy, zz ) ;
#endif sunos4

#ifdef sunos5
#define COPYIN( from, to, thing ) \
	copyin( (caddr_t)from, (caddr_t)to, sizeof( thing ) )
#endif sunos5

/*---------------------------------------------------------------------------*/
/* Interrupt lockout/restore macros...
*/
#ifdef sunos4
#define LOCKINT \
	s = splx( btfp_pritosplx( (int)md->md_intpri ) )
#endif sunos4

#ifdef sunos5
#define LOCKINT \
	mutex_enter( &sp->mutex )
#endif sunos5

#ifdef sunos4
#define UNLOCKINT \
	splx( s )
#endif sunos4

#ifdef sunos5
#define UNLOCKINT \
	mutex_exit( &sp->mutex )
#endif sunos5

/*---------------------------------------------------------------------------*/
/* 'Sleep/Wakeup' macros...
*/
#ifdef sunos4
#define INTWAIT( x ) \
	sleep( (caddr_t)&x, PZERO + 5 )
#endif sunos4

#ifdef sunos5
#define INTWAIT( x ) \
	cv_wait( &sp->cv, &sp->mutex )
#endif sunos5

#ifdef sunos4
#define WAITWAKEUP( x ) \
        wakeup( (caddr_t)&x )
#endif sunos4

#ifdef sunos5
#define WAITWAKEUP( x ) \
	cv_broadcast( &sp->cv )
#endif sunos5

/*---------------------------------------------------------------------------*/
