/****************************************************************************/
/*
	The include file for SunOS4 or SunOS5 applications calling 
		the bc635VME device driver.

	The device name is 'btfp' for 'BANCOMM time and frequency processor'.	

	Copyright (C) 1994, BANCOMM, San Jose, CA.  All rights reserved.	
                                                                      
  	Author:	Steve Schoggen

		S. G. Schoggen & Company
		P. O. Box 246							
		Little Silver, NJ  07739-0246		908-530-4659

		Winter, 1994

		01/21/94..SGS..To build an application for use with
				SunOS5, #define sunos5 before this
				routine.
				This file and the driver will default
				to building for SunOS4, so you need
				not do anything special when making 
				a SunOS4 kernel.
*/
/****************************************************************************/
#include <sys/ioccom.h>			/* Required for ioctl cmd definitions.*/
#include <sys/ioctl.h>

#ifndef sunos5				/* Compile for SunOS5.x.*/
#define sunos4				/* Compile for SunOS4.x.*/
#endif  sunos5

#ifdef sunos4
#define UNSIGNED unsigned
					/* SunOS4 ioctl commands definitions.*/
#define IOCIO    _IO
#define IOCIOR   _IOR
#define IOCIORN  _IORN
#define IOCIOWN  _IOWN
#endif sunos4

#ifdef sunos5
#define UNSIGNED volatile unsigned	/* Eliminate optimization problems.*/
					/* SunOS5 ioctl commands definitions.*/
#define BTFPIOC            ( 'b'<< 8 )
#define IOCIO( l, n )      ( BTFPIOC | n )
#define IOCIOR( l, n, s )  ( BTFPIOC | n )
#define IOCIORN( l, n, s ) ( BTFPIOC | n )
#define IOCIOWN( l, n, s ) ( BTFPIOC | n )
#endif sunos5

/****************************************************************************/
struct btfp_device			/* bc635 module register definitions.*/
					/* Note:  UNSIGNED short must be used!*/
{
	UNSIGNED short	btfp_id;		/* ID register.	*/
#define BC635ID		0xfef4			/* What this reg should read.*/
	UNSIGNED short	btfp_device;		/* Device type.*/
#define btfp_type	btfp_device		/* A closer name than device!*/
#define BC635TYPE	0xf350			/* What this reg should read.*/
	UNSIGNED short	btfp_status;		/* Status register.*/
#define btfp_control	btfp_status 		/* Control register.*/
	UNSIGNED short	btfp_gap1[ 2 ];		/* Unused address space.*/
	UNSIGNED short	btfp_timereq;		/* Time Request.*/
	UNSIGNED short	btfp_time [5];		/* Requested Time registers.*/
						/* XX, XX, Stat, DH */
						/* DT, DU, HT, HU */
						/* MT, MU, ST, SU */
						/* MSH, MST, MSU, USH */
						/* UST, USU, NSH, XX */
	UNSIGNED short	btfp_event [5];		/* Requested Event registers.*/
						/* XX, XX, Stat, DH */
						/* MT, MU, ST, SU */
						/* MSH, MST, MSU, USH */
						/* UST, USU, NSH, XX */

#define btfp_strobe1	btfp_event[1] 		/* Strobe part 1 */
						/* XX, XX, HT, HU */
#define btfp_strobe2	btfp_event[2] 		/* Strobe part 2 */
						/* MT, MU, ST, SU */
#define btfp_strobe3	btfp_event[3] 		/* Strobe part 3 */
						/* MSH, MST, MSU, XX */
	
	UNSIGNED short	btfp_unlock;		/* Release Capture Lockout*/
	UNSIGNED short	btfp_ack;		/* Data Acknowledge */
	UNSIGNED short	btfp_cr0;		/* Control Register 0*/
	UNSIGNED char	btfp_fifo[2];		/* I/O FIFO Data*/
#define btfp_outfifo	btfp_fifo[1]		/* Output FIFO (from module) */
#define btfp_infifo		btfp_fifo[1]	/* Input FIFO (to module) */
	UNSIGNED short	btfp_mask;		/* Interrupt Mask*/
	UNSIGNED short	btfp_intstat;		/* Interrupt Status*/
	UNSIGNED short	btfp_vector;		/* Interrupt Vector*/
#define BTFPVECTOR 0xc9				/* Hard-coded interrupt vector*/
	UNSIGNED short	btfp_level;		/* Interrupt Level*/
#define BTFPINTLEVEL 3				/* Hard-coded interrupt level.*/
	UNSIGNED short	btfp_gap2[ 6 ];		/* Unused address space.*/

}; /* end of btfp_device structure */
	
/****************************************************************************/
struct btfp_time				/* For reading 5 time words */
						/* in one ioctl(2) operation.*/
{
	UNSIGNED short	btfp_time[ 5 ];		/* Time words 0, 1, 2, 3, 4.. */
						/* (16bit).*/
};

/****************************************************************************/
struct btfp_event				/* For reading 5 event words */
						/* in one ioctl(2) operation.*/
{
	UNSIGNED short	btfp_event[ 5 ];	/* Event words 0, 1, 2, 3, 4..*/
						/* (16bit).*/
};

/****************************************************************************/
struct btfp_strobe				/* For writing 3 strobe words */
						/* in one ioctl(2) operation. */
{
	UNSIGNED short	btfp_strobe[ 3 ];	/* Strobe words 1, 2, and 3. */
						/* (16bit).*/
};

/****************************************************************************/
                 				/* 'I/O Control' commands.*/
                 				/* IOCIO,IOCIOR,&IOCIOW macros*/
						/* in /usr/sys/sys/ioccom.h.*/
                 				/* With IOCIO, arg is an int, */
						/* with the others it's a ptr!*/

/***** Simple (i.e. single-action) ioctl commands *****/


#define RIDR      	IOCIOR(b, 1, int )     	/* Read the ID register. */
#define RDTR      	IOCIOR(b, 2, int )  	/* Read Device Type register. */

#define RSR      	IOCIOR(b, 3, int )  	/* Read Status register. */
#define WCR      	IOCIO(b, 4)  		/* Write Control register. */

#define RTIMEREQ    	IOCIOR(b, 5, int )  	/* Read Time Request register.*/
#define RTIME0      	IOCIOR(b, 6, int )  	/* Read time word zero. */
#define RTIME1      	IOCIOR(b, 7, int )  	/* Read time word one. */
#define RTIME2      	IOCIOR(b, 8, int )  	/* Read time word two. */
#define RTIME3      	IOCIOR(b, 9, int )  	/* Read time word three. */
#define RTIME4      	IOCIOR(b, 10, int )  	/* Read time word four. */

#define REVENT0     	IOCIOR(b, 11, int )  	/* Read event word zero. */
#define REVENT1    	IOCIOR(b, 12, int )  	/* Read event word one. */
#define REVENT2     	IOCIOR(b, 13, int )	/* Read event word two. */
#define REVENT3     	IOCIOR(b, 14, int )  	/* Read event word three. */
#define REVENT4     	IOCIOR(b, 15, int )  	/* Read event word four. */

#define WSTROBE1    	IOCIO(b, 16)  		/* Write strobe word one. */
#define WSTROBE2    	IOCIO(b, 17)  		/* Write strobe word two. */
#define WSTROBE3    	IOCIO(b, 18)  		/* Write strobe word three. */

#define RUNLOCK     	IOCIOR(b, 19, int )  	/* Release Capture Lockout */

#define RACK      	IOCIOR(b, 20, int )  	/* Data Acknowledge.*/
#define WACK      	IOCIO(b, 21)  		/* Data Acknowledge.*/

#define RCR0      	IOCIOR(b, 22, int )  	/* Read control register zero.*/
#define	WCR0		IOCIO (b, 23)		/* Write control register zero*/
#define	ROUTFIFO	IOCIOR(b, 24, int ) 	/* Read from OUTFIFO. */
#define	WINFIFO		IOCIO(b, 25)		/* Write to INFIFO. */

#define RMASK      	IOCIOR(b, 26, int )  	/* Read Interrupt Mask. */
#define WMASK      	IOCIO(b, 27)  		/* Write Interrupt Mask. */

#define RINTSTAT    	IOCIOR(b, 28, int )  	/* Interrupt Status. */
#define WINTSTAT    	IOCIO(b, 29)  		/* Interrupt Status. */

#define RVECTOR     	IOCIOR(b, 30, int )  	/* Interrupt Vector. */
						/* Only driver can write to it*/

#define RLEVEL      	IOCIOR(b, 31, int )  	/* Interrupt Level. */
						/* Only driver can write to it*/


/***** Compound (i.e. multiple-action) ioctl commands *****/

					/* Read all 5 time words in one call.*/
#define READTIME	IOCIORN(b, 32, sizeof( struct btfp_time ))
				 	/* Read all 5 event words in 1 call.*/
#define READEVENT	IOCIORN(b, 33, sizeof( struct btfp_event ))
					/* Write all 3 strobe words in 1 call.*/
#define WRITESTROBE 	IOCIOWN(b, 34, sizeof( struct btfp_strobe ))
				 	/* Read all btfp regs.	*/
#define	READALLREGS	IOCIORN(b, 35, sizeof( struct btfp_device )) 
					/* Read packet from OUTFIFO */
#ifdef OMIT
					/* Not implemented as of now.*/
#define	RPACKET	 	IOCIORN(b, 36, sizeof( struct btfp_packet ))
					/* Write packet to INFIFO */
#define	WPACKET	 	IOCIOWN(b, 37, sizeof( struct btfp_packet )) 
#endif OMIT

#define SIGNALON	IOCIO(b, 38)	/* Turn on signal sending from driver.*/
					/* Also turns off blocking on intrrpts*/
#define SIGNALOFF	IOCIO(b, 39)	/* Turn off signal sending frm driver.*/
					/* Also turns on blocking on intrpts.*/
#define PACKETON	IOCIO(b, 40)	/* Turn on packet-mode for driver */
#define PACKETOFF	IOCIO(b, 41)	/* Turn off packet-mode for driver */
#define DEBUGON		IOCIO(b, 42)	/* Turn on debug-mode for driver */
#define DEBUGOFF	IOCIO(b, 43)	/* Turn off debug-mode for driver */
									 
/*****************************************************************************/
		/* bc635 Requested Time/Event register masks.*/

#define	TIME_REF	0x0000		/* Reference source present */
#define	EVENT_REF	0x0000		/* Reference source present */

#define	TIME_FLY	0x0010		/* Flywheeling; time code lost */
#define	EVENT_FLY	0x0010		/* Flywheeling; time code lost */

#define TIME_OSCIN	0x0000		/* Oscillator in range */
#define EVENT_OSCIN	0x0000		/* Oscillator in range */

#define TIME_OSCOUT	0x0040		/* Oscillator out of range */
#define EVENT_OSCOUT	0x0040		/* Oscillator out of range */

/****************************************************************************/
		/* bc635 ACK register masks */

#define ACK_INFIFO	0x0001		/* Input FIFO ACK Bit */
#define ACK_1PPS	0x0002		/* 1 Pulse Per Second ACK Bit */
#define ACK_DPA		0x0004		/* Output FIFO Data Pkt Available Bit */
#define ACK_MORE	0x0010		/* Output FIFO "More in OUTFIFO" Flag */
					/* 0: empty, 1: more bytes */
#define ACK_CLEAR	0x0010		/* Output FIFO Clear Bit */
#define ACK_INACT	0x0080		/* Input FIFO "Take Action" Bit */
/****************************************************************************/
		/* bc635 CR0 control register masks.*/

#define	LOCKEN		0x0001		/* Enable Capture Lockout */
#define	LOCKDIS		0x0000		/* Disable Capture Lockout */

#define HBEN		0x0002		/* Enable Periodic Time Capture */
#define HBDIS		0x0000		/* Disable Periodic Time Capture */

#define EVSENSE_FALL	0x0004		/* Event Input Active Edge Select */
					/* (Falling Edge)
#define EVSENSE_RISE	0x0000		/* Event Input Active Edge Select */
					/* (Rising Edge) */

#define EVENTEN		0x0008		/* Event Input Time Capture Enable */
#define EVENTDIS	0x0000		/* Event Input Time Capture Disable */

#define STREN		0x0010		/* Time Coincidence Strobe Out Enable*/
#define STRDIS		0x0000		/* Time Coincidence Strobe Out Disable*/

#define STRMODE_MIN	0x0020		/* Time Coincidence Strobe Mode Select*/
					/* (Minor-only mode) */
#define STRMODE_MAJ	0x0000		/* Time Coincidence Strobe Mode Select*/
					/* (Major/Minor mode) */

#define FREQSEL_10	0x0000		/* Output Clock 10 MHz */
#define FREQSEL_5	0x0040		/* Output Clock 5 MHz */
#define FREQSEL_1	0x0080		/* Output Clock 1 MHz */

/****************************************************************************/
		/* bc635 Interrupt Mask register masks.*/

#define MASK_INT0	0x0001		/* Interrupt Source 0 */
#define MASK_EXT	0x0001		/* External Event Input */

#define MASK_INT1	0x0002		/* Interrupt Source 1 */
#define MASK_HEART	0x0002		/* Periodic Pulse Output */

#define MASK_INT2	0x0004		/* Interrupt Source 2 */
#define MASK_STROBE	0x0004		/* Time Coincidence Strobe */

#define MASK_INT3	0x0008		/* Interrupt Source 3 */
#define MASK_1PPS	0x0008		/* 1 Pulse per Second (1PPS) */

#define MASK_INT4	0x0010		/* Interrupt Source 4 */
#define MASK_DPA	0x0010		/* Data Packet Available */

/****************************************************************************/
		/* bc635 Interrupt Status register masks.*/

#define INTSTAT_INT0	0x0001		/* Interrupt Source 0 */
#define INTSTAT_EXT	0x0001		/* External Event Input */

#define INTSTAT_INT1	0x0002		/* Interrupt Source 1 */
#define INTSTAT_HEART	0x0002		/* Periodic Pulse Output */

#define INTSTAT_INT2	0x0004		/* Interrupt Source 2 */
#define INTSTAT_STROBE	0x0004		/* Time Coincidence Strobe */

#define INTSTAT_INT3	0x0008		/* Interrupt Source 3 */
#define INTSTAT_1PPS	0x0008		/* 1 Pulse per Second (1PPS) */

#define INTSTAT_INT4	0x0010		/* Interrupt Source 4 */
#define INTSTAT_DPA	0x0010		/* Data Packet Available */

/****************************************************************************/
		/* bc635 Interrupt Level register masks */

#define LEVEL_DIS	0x0000		/* Interrupts disabled */
#define LEVEL_IRQ1	0x0001		/* IRQ1 */
#define LEVEL_IRQ2	0x0002		/* IRQ2 */
#define LEVEL_IRQ3	0x0003		/* IRQ3 */
#define LEVEL_IRQ4	0x0004		/* IRQ4 */
#define LEVEL_IRQ5	0x0005		/* IRQ5 */
#define LEVEL_IRQ6	0x0006		/* IRQ6 */
#define LEVEL_IRQ7	0x0007		/* IRQ7 */

/****************************************************************************/
		/* General defines */
#define	YES		1
#define NO		0
#define SOH		0x01
#define ETB		0x17
#define DLE 		0x10
#define ETX 		0x03
#define MAXPACKET	512		/* Maximum bytes in FIFO packet */
#define MAX_SPIN	100000		/* Max loops to chk for INFIFO action.*/

/****************************************************************************/
